"""Helper functions to save charts to a file, if plt.show() is called."""
import ast
import logging
import os
from os.path import dirname

import astor


def is_show_node(node: ast.Call) -> bool:
    if not hasattr(node, "value"):
        return False

    value = node.value
    return (
        isinstance(value, ast.Call)
        and value.func
        and (
            isinstance(value.func, ast.Attribute)
            and value.func.attr == "show"
            and value.func.value.id == "plt"
        )
    )


def add_save_chart(
    code: str,
    logger: logging.Logger,
    folder_name: str,
    save_charts_path: str = None,
) -> str:
    """
    Add line to code that save charts to a file, if plt.show() is called.

    Args:
        code (str): Code to add line to.
        logger (logging.Logger): Logger to use.
        folder_name (str): Name of folder to save charts to.
        save_charts_path (str): User Defined Path to save Charts

    Returns:
        str: Code with line added.

    """

    if save_charts_path is not None:
        charts_root_dir = save_charts_path
    else:
        # define chart save directory
        charts_root_dir = dirname(dirname(dirname(__file__)))

    chart_save_dir = os.path.join(charts_root_dir, "exports", "charts", folder_name)

    tree = ast.parse(code)

    # count number of plt.show() calls
    show_count = sum(1 for node in ast.walk(tree) if is_show_node(node))

    # if there are no plt.show() calls, return the original code
    if show_count == 0:
        return code

    if not os.path.exists(chart_save_dir):
        os.makedirs(chart_save_dir)

    # iterate through the AST and add plt.savefig() calls before plt.show() calls
    counter = ord("a")
    new_body = []
    for node in tree.body:
        if is_show_node(node):
            filename = "chart"
            if show_count > 1:
                filename += f"_{chr(counter)}"
                counter += 1

            chart_save_path = os.path.join(chart_save_dir, f"{filename}.png")
            new_body.append(ast.parse(f"plt.savefig(r'{chart_save_path}')"))
        new_body.append(node)

    chart_save_msg = f"Charts saving to: {chart_save_dir}"
    if not logger.verbose:
        print(chart_save_msg)
    logging.info(chart_save_msg)

    new_tree = ast.Module(body=new_body)
    return astor.to_source(new_tree, pretty_source=lambda x: "".join(x)).strip()



"""Unit tests for the CodeManager class"""
from typing import Optional
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from predicatepandasai.exceptions import BadImportError, NoCodeFoundError
from predicatepandasai.llm.fake import FakeLLM

from predicatepandasai.smart_dataframe import SmartDataframe

from predicatepandasai.helpers.code_manager import CodeManager


class TestCodeManager:
    """Unit tests for the CodeManager class"""

    @pytest.fixture
    def llm(self, output: Optional[str] = None):
        return FakeLLM(output=output)

    @pytest.fixture
    def sample_df(self):
        return pd.DataFrame(
            {
                "country": [
                    "United States",
                    "United Kingdom",
                    "France",
                    "Germany",
                    "Italy",
                    "Spain",
                    "Canada",
                    "Australia",
                    "Japan",
                    "China",
                ],
                "gdp": [
                    19294482071552,
                    2891615567872,
                    2411255037952,
                    3435817336832,
                    1745433788416,
                    1181205135360,
                    1607402389504,
                    1490967855104,
                    4380756541440,
                    14631844184064,
                ],
                "happiness_index": [
                    6.94,
                    7.16,
                    6.66,
                    7.07,
                    6.38,
                    6.4,
                    7.23,
                    7.22,
                    5.87,
                    5.12,
                ],
            }
        )

    @pytest.fixture
    def smart_dataframe(self, llm, sample_df):
        return SmartDataframe(sample_df, config={"llm": llm, "enable_cache": False})

    @pytest.fixture
    def code_manager(self, smart_dataframe: SmartDataframe):
        return smart_dataframe._dl._code_manager

    def test_run_code_for_calculations(self, code_manager: CodeManager):
        code = """def analyze_data(dfs):
    return {'type': 'number', 'value': 1 + 1}"""

        assert code_manager.execute_code(code, "")["value"] == 2
        assert code_manager.last_code_executed == code

    def test_run_code_invalid_code(self, code_manager: CodeManager):
        with pytest.raises(Exception):
            code_manager.execute_code("1+ ", "")

    def test_clean_code_remove_builtins(self, code_manager: CodeManager):
        builtins_code = """import set
def analyze_data(dfs):
    return {'type': 'number', 'value': set([1, 2, 3])}"""
        assert code_manager.execute_code(builtins_code, "")["value"] == {1, 2, 3}
        assert (
            code_manager.last_code_executed
            == """def analyze_data(dfs):
    return {'type': 'number', 'value': set([1, 2, 3])}"""
        )

    def test_clean_code_removes_jailbreak_code(self, code_manager: CodeManager):
        malicious_code = """def analyze_data(dfs):
    __builtins__['str'].__class__.__mro__[-1].__subclasses__()[140].__init__.__globals__['system']('ls')
    print('hello world')"""
        assert (
            code_manager._clean_code(malicious_code)
            == """def analyze_data(dfs):
    print('hello world')"""
        )

    def test_clean_code_remove_environment_defaults(self, code_manager: CodeManager):
        pandas_code = """import pandas as pd
print('hello world')
"""
        assert code_manager._clean_code(pandas_code) == "print('hello world')"

    def test_clean_code_whitelist_import(self, code_manager: CodeManager):
        """Test that an installed whitelisted library is added to the environment."""
        safe_code = """
import numpy as np
np.array()
"""
        assert code_manager._clean_code(safe_code) == "np.array()"

    def test_clean_code_raise_bad_import_error(self, code_manager: CodeManager):
        malicious_code = """
import os
print(os.listdir())
"""
        with pytest.raises(BadImportError):
            code_manager.execute_code(malicious_code, "")

    def test_remove_dfs_overwrites(self, code_manager: CodeManager):
        hallucinated_code = """def analyze_data(dfs):
    dfs = [pd.DataFrame([1,2,3])]
    print(dfs)"""
        assert (
            code_manager._clean_code(hallucinated_code)
            == """def analyze_data(dfs):
    print(dfs)"""
        )

    def test_exception_handling(
        self, smart_dataframe: SmartDataframe, code_manager: CodeManager
    ):
        code_manager.execute_code = Mock(
            side_effect=NoCodeFoundError("No code found in the answer.")
        )

        result = smart_dataframe.chat("How many countries are in the dataframe?")
        assert result == (
            "Unfortunately, I was not able to answer your question, "
            "because of the following error:\n"
            "\nNo code found in the answer.\n"
        )
        assert smart_dataframe.last_error == "No code found in the answer."

    def test_custom_whitelisted_dependencies(self, code_manager: CodeManager, llm):
        code = """
import my_custom_library
def analyze_data(dfs: list):
    my_custom_library.do_something()
"""
        llm._output = code

        with pytest.raises(BadImportError):
            code_manager._clean_code(code)

        code_manager._config.custom_whitelisted_dependencies = ["my_custom_library"]
        assert (
            code_manager._clean_code(code)
            == """def analyze_data(dfs: list):
    my_custom_library.do_something()"""
        )

    def test_get_environment(self, code_manager: CodeManager, smart_dataframe):
        code_manager._additional_dependencies = [
            {"name": "pyplot", "alias": "plt", "module": "matplotlib"},
            {"name": "numpy", "alias": "np", "module": "numpy"},
        ]

        assert smart_dataframe.equals(code_manager._get_environment()["dfs"][0])
        assert "pd" in code_manager._get_environment()
        assert "plt" in code_manager._get_environment()
        assert "np" in code_manager._get_environment()
        assert code_manager._get_environment()["__builtins__"] == {
            "abs": abs,
            "all": all,
            "any": any,
            "ascii": ascii,
            "bin": bin,
            "bool": bool,
            "bytearray": bytearray,
            "bytes": bytes,
            "callable": callable,
            "chr": chr,
            "classmethod": classmethod,
            "complex": complex,
            "delattr": delattr,
            "dict": dict,
            "dir": dir,
            "divmod": divmod,
            "enumerate": enumerate,
            "filter": filter,
            "float": float,
            "format": format,
            "frozenset": frozenset,
            "getattr": getattr,
            "hasattr": hasattr,
            "hash": hash,
            "help": help,
            "hex": hex,
            "id": id,
            "int": int,
            "isinstance": isinstance,
            "issubclass": issubclass,
            "iter": iter,
            "len": len,
            "list": list,
            "locals": locals,
            "map": map,
            "max": max,
            "memoryview": memoryview,
            "min": min,
            "next": next,
            "object": object,
            "oct": oct,
            "ord": ord,
            "pow": pow,
            "print": print,
            "property": property,
            "range": range,
            "repr": repr,
            "reversed": reversed,
            "round": round,
            "set": set,
            "setattr": setattr,
            "slice": slice,
            "sorted": sorted,
            "staticmethod": staticmethod,
            "str": str,
            "sum": sum,
            "super": super,
            "tuple": tuple,
            "type": type,
            "vars": vars,
            "zip": zip,
            "__build_class__": __build_class__,
            "__name__": "__main__",
        }

    def test_execute_catching_errors_correct(self, code_manager: CodeManager):
        code = """def analyze_data(dfs):
    return {'type': 'number', 'value': 1 + 1}"""
        environment = {"dfs": []}

        with patch("builtins.exec") as mock_exec:
            assert code_manager._execute_catching_errors(code, environment) is None
            mock_exec.assert_called_once_with(
                code + "\n\nresult = analyze_data(dfs)", environment
            )

    def test_execute_catching_errors_raise_exc(self, code_manager: CodeManager):
        code = """def analyze_data(dfs):
    raise RuntimeError()"""
        environment = {"dfs": []}

        with patch("builtins.exec") as mock_exec:
            mock_exec.side_effect = RuntimeError("foobar")
            exc = code_manager._execute_catching_errors(code, environment)
            mock_exec.assert_called_once_with(
                code + "\n\nresult = analyze_data(dfs)", environment
            )
            assert isinstance(exc, RuntimeError)

    def test_handle_error_name_error(self, code_manager: CodeManager):
        code = """def analyze_data(dfs):
    print(json.dumps({"foo": "bar"}))"""
        environment = {"dfs": []}

        exc = code_manager._execute_catching_errors(code, environment)
        code_manager._handle_error(exc, code, environment)
        assert getattr(environment.get("json"), "__name__", None) == "json"

    def test_handle_error_name_error_not_whitelisted_lib(
        self, code_manager: CodeManager
    ):
        code = """def analyze_data(dfs):
    print(os)"""
        environment = {"dfs": []}

        exc = code_manager._execute_catching_errors(code, environment)
        with pytest.raises(NameError):
            code_manager._handle_error(
                exc, code, environment, use_error_correction_framework=False
            )
        assert "os" not in environment



"""Fake LLM"""

from typing import Optional

from ..prompts.base import Prompt
from .base import LLM


class FakeLLM(LLM):
    """Fake LLM"""

    _output: str = """def analyze_data(dfs):
    return { 'type': 'text', 'value': "Hello World" }"""

    def __init__(self, output: Optional[str] = None):
        if output is not None:
            self._output = output

    def call(self, instruction: Prompt, suffix: str = "") -> str:
        self.last_prompt = instruction.to_string() + suffix
        return self._output

    @property
    def type(self) -> str:
        return "fake"



""" Falcon LLM
This module is to run the Falcon API hosted and maintained by HuggingFace.co.
To generate HF_TOKEN go to https://huggingface.co/settings/tokens after creating Account
on the platform.

Example:
    Use below example to call Falcon Model

    >>> from predicatepandasai.llm.falcon import Falcon
"""


from ..helpers import load_dotenv
from .base import HuggingFaceLLM

load_dotenv()


class Falcon(HuggingFaceLLM):

    """Falcon LLM API

    A base HuggingFaceLLM class is extended to use Falcon model.

    """

    api_token: str
    _api_url: str = (
        "https://api-inference.huggingface.co/models/tiiuae/falcon-7b-instruct"
    )
    _max_retries: int = 30

    @property
    def type(self) -> str:
        return "falcon"



"""Unit tests for the SmartDatalake class"""
import os

from typing import Optional
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from predicatepandasai import SmartDataframe, SmartDatalake
from predicatepandasai.helpers.code_manager import CodeManager
from predicatepandasai.llm.fake import FakeLLM
from predicatepandasai.middlewares import Middleware

from langchain import OpenAI


class TestSmartDatalake:
    """Unit tests for the SmartDatlake class"""

    @pytest.fixture
    def llm(self, output: Optional[str] = None):
        return FakeLLM(output=output)

    @pytest.fixture
    def sample_df(self):
        return pd.DataFrame(
            {
                "country": [
                    "United States",
                    "United Kingdom",
                    "France",
                    "Germany",
                    "Italy",
                    "Spain",
                    "Canada",
                    "Australia",
                    "Japan",
                    "China",
                ],
                "gdp": [
                    19294482071552,
                    2891615567872,
                    2411255037952,
                    3435817336832,
                    1745433788416,
                    1181205135360,
                    1607402389504,
                    1490967855104,
                    4380756541440,
                    14631844184064,
                ],
                "happiness_index": [
                    6.94,
                    7.16,
                    6.66,
                    7.07,
                    6.38,
                    6.4,
                    7.23,
                    7.22,
                    5.87,
                    5.12,
                ],
            }
        )

    @pytest.fixture
    def smart_dataframe(self, llm, sample_df):
        return SmartDataframe(sample_df, config={"llm": llm, "enable_cache": False})

    @pytest.fixture
    def smart_datalake(self, smart_dataframe: SmartDataframe):
        return smart_dataframe.datalake

    @pytest.fixture
    def custom_middleware(self):
        class CustomMiddleware(Middleware):
            def run(self, code):
                return """def analyze_data(dfs):
    return { 'type': 'text', 'value': "Overwritten by middleware" }"""

        return CustomMiddleware

    def test_load_llm_with_pandasai_llm(self, smart_datalake: SmartDatalake, llm):
        smart_datalake._llm = None
        assert smart_datalake._llm is None

        smart_datalake._load_llm(llm)
        assert smart_datalake._llm == llm

    def test_load_llm_with_langchain_llm(self, smart_datalake: SmartDatalake, llm):
        langchain_llm = OpenAI(openai_api_key="fake_key")

        smart_datalake._llm = None
        assert smart_datalake._llm is None

        smart_datalake._load_llm(langchain_llm)
        assert smart_datalake._llm._langchain_llm == langchain_llm

    @patch.object(
        CodeManager,
        "execute_code",
        return_value={
            "type": "string",
            "value": "There are 10 countries in the dataframe.",
        },
    )
    def test_last_result_is_saved(self, _mocked_method, smart_datalake: SmartDatalake):
        assert smart_datalake.last_result is None

        smart_datalake.chat("How many countries are in the dataframe?")
        assert smart_datalake.last_result == {
            "type": "string",
            "value": "There are 10 countries in the dataframe.",
        }

    def test_middlewares(self, smart_dataframe: SmartDataframe, custom_middleware):
        middleware = custom_middleware()
        smart_dataframe._dl._code_manager._middlewares = [middleware]
        assert smart_dataframe._dl.middlewares == [middleware]
        assert (
            smart_dataframe.chat("How many countries are in the dataframe?")
            == "Overwritten by middleware"
        )
        assert middleware.has_run

    def test_retry_on_error_with_single_df(
        self, smart_datalake: SmartDatalake, smart_dataframe: SmartDataframe
    ):
        code = """def analyze_data(df):
    return { "type": "text", "value": "Hello World" }"""

        smart_dataframe._get_head_csv = Mock(
            return_value="""country,gdp,happiness_index
China,0654881226,6.66
Japan,9009692259,7.16
Spain,8446903488,6.38"""
        )

        smart_datalake._retry_run_code(
            code=code,
            e=Exception("Test error"),
        )

        assert (
            smart_datalake.last_prompt
            == """
You are provided with a pandas dataframe (df) with 10 rows and 3 columns.
This is the metadata of the dataframe:
country,gdp,happiness_index
China,0654881226,6.66
Japan,9009692259,7.16
Spain,8446903488,6.38.

The user asked the following question:


You generated this python code:
def analyze_data(df):
    return { "type": "text", "value": "Hello World" }

It fails with the following error:
Test error

Correct the python code and return a new python code (do not import anything) that fixes the above mentioned error. Do not generate the same code again.
"""  # noqa: E501
        )

    @patch("os.makedirs")
    def test_initialize(self, mock_makedirs, smart_datalake: SmartDatalake):
        smart_datalake.initialize()

        charts_dir = os.path.join(os.getcwd(), "exports", "charts")
        mock_makedirs.assert_any_call(charts_dir, mode=0o777, exist_ok=True)

        cache_dir = os.path.join(os.getcwd(), "cache")
        mock_makedirs.assert_any_call(cache_dir, mode=0o777, exist_ok=True)



"""
Charts Middleware class

Middleware to handle the charts in PandasAI.
"""

from predicatepandasai.middlewares.base import Middleware
import sys
import ast


class ChartsMiddleware(Middleware):
    """Charts Middleware class"""

    def _is_running_in_console(self) -> bool:
        """
        Check if the code is running in console or not.

        Returns:
            bool: True if running in console else False
        """

        return sys.stdout.isatty()

    def run(self, code: str) -> str:
        """
        Run the middleware to remove issues with displaying charts in PandasAI.

        Returns:
            str: Modified code
        """

        tree = ast.parse(code)

        for node in ast.walk(tree):
            if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
                call = node.value
                if isinstance(call.func, ast.Attribute) and call.func.attr == "show":
                    if "plt.close('all')" not in code:
                        code = code.replace(
                            "plt.show()", "plt.show()\nplt.close('all')"
                        )

                    if not self._is_running_in_console():
                        code = code.replace("plt.show()", "plt.show(block=False)")
                    break
        return code



from .base import LLM, HuggingFaceLLM
from .azure_openai import AzureOpenAI
from .openai import OpenAI
from .falcon import Falcon
from .google_palm import GooglePalm, GoogleVertexai
from .langchain import LangchainLLM
from .starcoder import Starcoder

__all__ = [
    "LLM",
    "HuggingFaceLLM",
    "AzureOpenAI",
    "OpenAI",
    "Falcon",
    "GooglePalm",
    "GoogleVertexai",
    "LangchainLLM",
    "Starcoder",
]



""" Base class to implement a new Prompt
In order to better handle the instructions, this prompt module is written.
"""

from predicatepandasai.exceptions import MethodNotImplementedError
from string import Formatter


class Prompt:
    """Base class to implement a new Prompt"""

    text = None
    _args = {}

    def __init__(self, **kwargs):
        """
        __init__ method of Base class of Prompt Module
        Args:
            **kwargs: Inferred Keyword Arguments
        """
        if kwargs:
            self._args = kwargs

        """Set all the variables with underscore prefix with default value as DEFAULT
        This will prevent any possible key errors if anyone tries to print 
        prompt before running .run method"""
        if self.text:
            vars_ = [
                fn for _, fn, _, _ in Formatter().parse(self.text) if fn is not None
            ]
            for var in vars_:
                if var[0] == "_" and var not in self._args:
                    self._args[var] = var

    def override_var(self, var, value):
        self._args[var] = value

    def to_string(self):
        if self.text is None:
            raise MethodNotImplementedError

        return self.text.format(**self._args)

    def __str__(self):
        return self.to_string()



# -*- coding: utf-8 -*-
"""
PandasAI is a wrapper around a LLM to make dataframes conversational

This module includes the implementation of basis  PandasAI class with methods to run
the LLMs models on Pandas dataframes. Following LLMs are implemented so far.

Example:

    This module is the Entry point of the `pandasai` package. Following is an example
    of how to use this Class.

    ```python
    import pandas as pd
    from predicatepandasai import PandasAI

    # Sample DataFrame
    df = pd.DataFrame({
        "country": ["United States", "United Kingdom", "France", "Germany", "Italy",
        "Spain", "Canada", "Australia", "Japan", "China"],
        "gdp": [19294482071552, 2891615567872, 2411255037952, 3435817336832,
        1745433788416, 1181205135360, 1607402389504, 1490967855104, 4380756541440,
        14631844184064],
        "happiness_index": [6.94, 7.16, 6.66, 7.07, 6.38, 6.4, 7.23, 7.22, 5.87, 5.12]
    })

    # Instantiate a LLM
    from predicatepandasai.llm.openai import OpenAI
    llm = OpenAI(api_token="YOUR_API_TOKEN")

    pandas_ai = PandasAI(llm)
    pandas_ai(df, prompt='Which are the 5 happiest countries?')

    ```
"""

from typing import List, Optional, Union, Dict, Type
import importlib.metadata

import pandas as pd
from .smart_dataframe import SmartDataframe
from .smart_datalake import SmartDatalake
from .prompts.base import Prompt
from .callbacks.base import BaseCallback
from .helpers.df_config import Config
from .helpers.cache import Cache

__version__ = importlib.metadata.version(__package__ or __name__)


class PandasAI:
    """
    PandasAI is a wrapper around a LLM to make dataframes conversational.


    This is an entry point of `pandasai` object. This class consists of methods
    to interface the LLMs with Pandas     dataframes. A pandas dataframe metadata i.e.
    df.head() and prompt is passed on to chosen LLMs API end point to generate a Python
    code to answer the questions asked. The resultant python code is run on actual data
    and answer is converted into a conversational form.

    Note:
        Do not include the `self` parameter in the ``Args`` section.
    Args:
        _llm (obj): LLMs option to be used for API access
        _verbose (bool, optional): To show the intermediate outputs e.g. python code
        generated and execution step on the prompt. Default to False
        _enforce_privacy (bool, optional): Do not display the data on prompt in case of
        Sensitive data. Default to False
        _max_retries (int, optional): max no. of tries to generate code on failure.
        Default to 3
        _in_notebook (bool, optional): Whether to run code in notebook. Default to False
        _original_instructions (dict, optional): The dict of instruction to run. Default
        to None
        _cache (Cache, optional): Cache object to store the results. Default to None
        _enable_cache (bool, optional): Whether to enable cache. Default to True
        _logger (logging.Logger, optional): Logger object to log the messages. Default
        to None
        _logs (List[dict], optional): List of logs to be stored. Default to []
        _prompt_id (str, optional): Unique ID to differentiate calls. Default to None
        _middlewares (List[Middleware], optional): List of middlewares to run. Default
        to [ChartsMiddleware()]
        _additional_dependencies (List[dict], optional): List of additional dependencies
        to be added. Default to []
        _custom_whitelisted_dependencies (List[str], optional): List of custom
        whitelisted dependencies. Default to []
        last_code_generated (str, optional): Pass last Code if generated. Default to
        None
        last_code_executed (str, optional): Pass the last execution / run. Default to
        None
        code_output (str, optional): The code output if any. Default to None
        last_error (str, optional): Error of running code last time. Default to None
        prompt_id (str, optional): Unique ID to differentiate calls. Default to None


    Returns (str): Response to a Question related to Data

    """

    _dl: SmartDatalake
    _config: Config

    def __init__(
        self,
        llm=None,
        conversational=False,
        verbose=False,
        enforce_privacy=False,
        save_charts=False,
        save_charts_path="",
        enable_cache=True,
        middlewares=[],
        custom_whitelisted_dependencies=[],
        enable_logging=True,
        non_default_prompts: Optional[Dict[str, Type[Prompt]]] = None,
        callback: Optional[BaseCallback] = None,
    ):
        """
        __init__ method of the Class PandasAI

        Args:
            llm (object): LLMs option to be used for API access. Default is None
            conversational (bool): Whether to return answer in conversational form.
            Default to False
            verbose (bool): To show the intermediate outputs e.g. python code
            generated and execution step on the prompt.  Default to False
            enforce_privacy (bool): Execute the codes with Privacy Mode ON.
            Default to False
            save_charts (bool): Save the charts generated in the notebook.
            Default to False
            enable_cache (bool): Enable the cache to store the results.
            Default to True
            middlewares (list): List of middlewares to be used. Default to None
            custom_whitelisted_dependencies (list): List of custom dependencies to
            be used. Default to None
            enable_logging (bool): Enable the logging. Default to True
            non_default_prompts (dict): Mapping from keys to replacement prompt classes.
            Used to override specific types of prompts. Defaults to None.
        """

        # configure the logging
        # noinspection PyArgumentList
        # https://stackoverflow.com/questions/61226587/pycharm-does-not-recognize-logging-basicconfig-handlers-argument

        self._config = Config(
            conversational=conversational,
            verbose=verbose,
            enforce_privacy=enforce_privacy,
            save_charts=save_charts,
            save_charts_path=save_charts_path,
            enable_cache=enable_cache,
            middlewares=middlewares,
            custom_whitelisted_dependencies=custom_whitelisted_dependencies,
            enable_logging=enable_logging,
            non_default_prompts=non_default_prompts,
            llm=llm,
            callback=callback,
        )

    def run(
        self,
        data_frame: Union[pd.DataFrame, List[pd.DataFrame]],
        prompt: str,
        show_code: bool = False,
        anonymize_df: bool = True,
        use_error_correction_framework: bool = True,
    ) -> Union[str, pd.DataFrame]:
        """
        Run the PandasAI to make Dataframes Conversational.

        Args:
            data_frame (Union[pd.DataFrame, List[pd.DataFrame]]): A pandas Dataframe
            prompt (str): A prompt to query about the Dataframe
            show_code (bool): To show the intermediate python code generated on the
            prompt. Default to False
            anonymize_df (bool): Running the code with Sensitive Data. Default to True
            use_error_correction_framework (bool): Turn on Error Correction mechanism.
            Default to True

        Returns (str): Answer to the Input Questions about the DataFrame

        """

        new_config = self._config.dict()
        new_config["show_code"] = show_code
        new_config["anonymize_df"] = anonymize_df
        new_config["use_error_correction_framework"] = use_error_correction_framework

        config = Config(**new_config).dict()

        if not isinstance(data_frame, list):
            data_frame = [data_frame]

        self._dl = SmartDatalake(data_frame, config)
        return self._dl.chat(prompt)

    def __call__(
        self,
        data_frame: Union[pd.DataFrame, List[pd.DataFrame]],
        prompt: str,
        show_code: bool = False,
        anonymize_df: bool = True,
        use_error_correction_framework: bool = True,
    ) -> Union[str, pd.DataFrame]:
        """
        __call__ method of PandasAI class. It calls the `run` method.

        Args:
            data_frame:
            prompt:
            show_code:
            anonymize_df:
            use_error_correction_framework:

        Returns (str): Answer to the Input Questions about the DataFrame.

        """

        return self.run(
            data_frame,
            prompt,
            show_code,
            anonymize_df,
            use_error_correction_framework,
        )

    @property
    def logs(self) -> List[dict[str, str]]:
        """Return the logs"""
        return self._dl.logs

    @property
    def last_prompt_id(self) -> str:
        """Return the id of the last prompt that was run."""
        return self._dl.last_prompt_id

    @property
    def last_prompt(self) -> str:
        """Return the last prompt that was executed."""
        return self._dl.last_prompt


def clear_cache(filename: str = None):
    """Clear the cache"""
    cache = Cache(filename or "cache")
    cache.clear()


__all__ = ["PandasAI", "SmartDataframe", "SmartDatalake", "clear_cache"]



""" Memory class to store the conversations """


class Memory:
    """Memory class to store the conversations"""

    _messages: list

    def __init__(self):
        self._messages = []

    def add(self, message: str, is_user: bool):
        self._messages.append({"message": message, "is_user": is_user})

    def count(self) -> int:
        return len(self._messages)

    def all(self) -> list:
        return self._messages

    def last(self) -> dict:
        return self._messages[-1]

    def get_conversation(self, limit: int = 1) -> str:
        return "\n".join(
            [
                f"{'User' if message['is_user'] else 'Bot'}: {message['message']}"
                for message in self._messages[-limit:]
            ]
        )

    def clear(self):
        self._messages = []



import re
import ast
import astor
import pandas as pd
from .save_chart import add_save_chart
from .optional import import_dependency
from ..exceptions import BadImportError
from ..middlewares.base import Middleware
from ..constants import (
    WHITELISTED_BUILTINS,
    WHITELISTED_LIBRARIES,
)
from ..middlewares.charts import ChartsMiddleware
from typing import Union, List, Optional
from ..helpers.logger import Logger
from ..helpers.df_config import Config
import logging
import traceback


class CodeManager:
    _dfs: List
    _middlewares: List[Middleware] = [ChartsMiddleware()]
    _config: Config
    _logger: Logger = None
    _additional_dependencies: List[dict] = []

    _last_code_executed: str = None

    def __init__(
        self,
        dfs: List,
        config: Config,
        logger: Logger,
    ):
        """
        Args:
            config (Config, optional): Config to be used. Defaults to None.
            logger (Logger, optional): Logger to be used. Defaults to None.
        """

        self._dfs = dfs
        self._config = config
        self._logger = logger

        if self._config.middlewares is not None:
            self.add_middlewares(*self._config.middlewares)

    def add_middlewares(self, *middlewares: List[Middleware]):
        """
        Add middlewares to PandasAI instance.

        Args:
            *middlewares: A list of middlewares

        """
        self._middlewares.extend(middlewares)

    def _execute_catching_errors(
        self, code: str, environment: dict
    ) -> Optional[Exception]:
        """
        Perform execution of the code directly.
        Call `exec()` for the given `code`, catch any non-base exceptions.
        Args:
            code (str): Python code
            environment (dict): Context for the `exec()`
        Returns (Optional[Exception]): Any exception raised during execution.
                                       `None` if executed without exceptions.
        """
        try:
            # Check in the code that analyze_data function is called.
            # If not, add it.
            if " = analyze_data(" not in code:
                code += "\n\nresult = analyze_data(dfs)"

            exec(code, environment)
        except Exception as exc:
            self._logger.log("Error of executing code", level=logging.WARNING)
            self._logger.log(f"{traceback.format_exc()}", level=logging.DEBUG)

            return exc

    def _handle_error(
        self,
        exc: Exception,
        code: str,
        environment: dict,
        use_error_correction_framework: bool = True,
    ):
        """
        Handle error occurred during first executing of code.
        If `exc` is instance of `NameError`, try to import the name, extend
        the context and then call `_execute_catching_errors()` again.
        If OK, returns the code string; if failed, continuing handling.
        Args:
            exc (Exception): The caught exception.
            code (str): Python code.
            environment (dict): Context for the `exec()`
        Raises:
            Exception: Any exception which has been caught during
                       the very first execution of the code.
        Returns (str): Python code. Either an original or new, given by
                       error correction framework.
        """
        if isinstance(exc, NameError):
            name_to_be_imported = None
            if hasattr(exc, "name"):
                name_to_be_imported = exc.name
            elif exc.args and isinstance(exc.args[0], str):
                name_ptrn = r"'([0-9a-zA-Z_]+)'"
                if search_name_res := re.search(name_ptrn, exc.args[0]):
                    name_to_be_imported = search_name_res.group(1)

            if name_to_be_imported and name_to_be_imported in WHITELISTED_LIBRARIES:
                try:
                    package = import_dependency(name_to_be_imported)
                    environment[name_to_be_imported] = package

                    caught_error = self._execute_catching_errors(code, environment)
                    if caught_error is None:
                        return code

                except ModuleNotFoundError:
                    self._logger.log(
                        f"Unable to fix `NameError`: package '{name_to_be_imported}'"
                        f" could not be imported.",
                        level=logging.DEBUG,
                    )
                except Exception as new_exc:
                    exc = new_exc
                    self._logger.log(
                        f"Unable to fix `NameError`: an exception was raised: "
                        f"{traceback.format_exc()}",
                        level=logging.DEBUG,
                    )

            if not use_error_correction_framework:
                raise exc

    def execute_code(
        self,
        code: str,
        prompt_id: str,
    ) -> str:
        """
        Execute the python code generated by LLMs to answer the question
        about the input dataframe. Run the code in the current context and return the
        result.

        Args:
            code (str): Python code to execute
            data_frame (pd.DataFrame): Full Pandas DataFrame
            use_error_correction_framework (bool): Turn on Error Correction mechanism.
            Default to True

        Returns:
            result: The result of the code execution. The type of the result depends
            on the generated code.

        """

        for middleware in self._middlewares:
            code = middleware(code)

        # Add save chart code
        if self._config.save_charts:
            code = add_save_chart(
                code,
                logger=self._logger,
                folder_name=str(prompt_id),
                save_charts_path=self._config.save_charts_path,
            )

        # Get the code to run removing unsafe imports and df overwrites
        code_to_run = self._clean_code(code)
        self.last_code_executed = code_to_run
        self._logger.log(
            f"""
Code running:
```
{code_to_run}
        ```"""
        )

        environment: dict = self._get_environment()

        caught_error = self._execute_catching_errors(code_to_run, environment)
        if caught_error is not None:
            self._handle_error(
                caught_error,
                code_to_run,
                environment,
                use_error_correction_framework=self._config.use_error_correction_framework,
            )

        analyze_data = environment.get("analyze_data", None)

        return analyze_data(self._get_original_dfs())

    def _get_original_dfs(self):
        dfs = []
        for df in self._dfs:
            if df.engine == "polars":
                dfs.append(df.original.to_pandas())
            else:
                dfs.append(df.original)

        return dfs

    def _get_environment(self) -> dict:
        """
        Returns the environment for the code to be executed.

        Returns (dict): A dictionary of environment variables
        """

        dfs = self._get_original_dfs()

        return {
            "pd": pd,
            "dfs": dfs,
            **{
                lib["alias"]: getattr(import_dependency(lib["module"]), lib["name"])
                if hasattr(import_dependency(lib["module"]), lib["name"])
                else import_dependency(lib["module"])
                for lib in self._additional_dependencies
            },
            "__builtins__": {
                **{builtin: __builtins__[builtin] for builtin in WHITELISTED_BUILTINS},
                "__build_class__": __build_class__,
                "__name__": "__main__",
            },
        }

    def _is_jailbreak(self, node: ast.stmt) -> bool:
        """
        Remove jailbreaks from the code to prevent malicious code execution.
        Args:
            node (object): ast.stmt
        Returns (bool):
        """

        DANGEROUS_BUILTINS = ["__subclasses__", "__builtins__", "__import__"]

        node_str = ast.dump(node)

        for builtin in DANGEROUS_BUILTINS:
            if builtin in node_str:
                return True

        return False

    def _is_unsafe(self, node: ast.stmt) -> bool:
        """
        Remove unsafe code from the code to prevent malicious code execution.

        Args:
            node (object): ast.stmt

        Returns (bool):
        """

        code = astor.to_source(node)
        if any(
            method in code
            for method in [
                ".to_csv",
                ".to_excel",
                ".to_json",
                ".to_sql",
                ".to_feather",
                ".to_hdf",
                ".to_parquet",
                ".to_pickle",
                ".to_gbq",
                ".to_stata",
                ".to_records",
                ".to_string",
                ".to_latex",
                ".to_html",
                ".to_markdown",
                ".to_clipboard",
            ]
        ):
            return True

        return False

    def _sanitize_analyze_data(self, analyze_data_node: ast.stmt) -> ast.stmt:
        # Sanitize the code within analyze_data
        sanitized_analyze_data = []
        for node in analyze_data_node.body:
            if (
                self._is_df_overwrite(node)
                or self._is_jailbreak(node)
                or self._is_unsafe(node)
            ):
                continue
            sanitized_analyze_data.append(node)

        analyze_data_node.body = sanitized_analyze_data
        return analyze_data_node

    def _clean_code(self, code: str) -> str:
        """
        A method to clean the code to prevent malicious code execution

        Args:
            code(str): A python code

        Returns (str): Returns a Clean Code String

        """

        # Clear recent optional dependencies
        self._additional_dependencies = []

        tree = ast.parse(code)

        # Check for imports and the node where analyze_data is defined
        new_body = []
        for node in tree.body:
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                self._check_imports(node)
                continue
            if isinstance(node, ast.FunctionDef) and node.name == "analyze_data":
                analyze_data_node = node
                sanitized_analyze_data = self._sanitize_analyze_data(analyze_data_node)
                new_body.append(sanitized_analyze_data)
                continue
            new_body.append(node)

        new_tree = ast.Module(body=new_body)
        return astor.to_source(new_tree, pretty_source=lambda x: "".join(x)).strip()

    def _is_df_overwrite(self, node: ast.stmt) -> bool:
        """
        Remove df declarations from the code to prevent malicious code execution.

        Args:
            node (object): ast.stmt

        Returns (bool):

        """

        return (
            isinstance(node, ast.Assign)
            and isinstance(node.targets[0], ast.Name)
            and node.targets[0].id == "dfs"
        )

    def _check_imports(self, node: Union[ast.Import, ast.ImportFrom]):
        """
        Add whitelisted imports to _additional_dependencies.

        Args:
            node (object): ast.Import or ast.ImportFrom

        Raises:
            BadImportError: If the import is not whitelisted

        """
        if isinstance(node, ast.Import):
            module = node.names[0].name
        else:
            module = node.module

        library = module.split(".")[0]

        if library == "pandas":
            return

        if (
            library
            in WHITELISTED_LIBRARIES + self._config.custom_whitelisted_dependencies
        ):
            for alias in node.names:
                self._additional_dependencies.append(
                    {
                        "module": module,
                        "name": alias.name,
                        "alias": alias.asname or alias.name,
                    }
                )
            return

        if library not in WHITELISTED_BUILTINS:
            raise BadImportError(library)

    @property
    def middlewares(self):
        return self._middlewares

    @property
    def last_code_executed(self):
        return self._last_code_executed

    @last_code_executed.setter
    def last_code_executed(self, code: str):
        self._last_code_executed = code



""" Driver code for the CLI tool
Pai is the command line tool designed to provide a convenient way to interact with
PandasAI through a command line interface (CLI).

```
pai [OPTIONS]
```
Options:

- **-d, --dataset**: The file path to the dataset.
- **-t, --token**: Your HuggingFace or OpenAI API token, if no token provided pai will
pull from the `.env` file.
- **-m, --model**: Choice of LLM, either `openai`, `open-assistant`, `starcoder`,
`falcon`, `azure-openai` or `google-palm`.
- **-p, --prompt**: Prompt that PandasAI will run.

To view a full list of available options and their descriptions, run the following
command:

```
pai --help

```

Example:
    ```
    pai -d "~/pandasai/example/data/Loan payments data.csv" -m "openai"
    -p "How many loans are from men and have been paid off?"

    ```

> Should result in the same output as the `from_csv.py` example.

"""
import os

import click
import pandas as pd

from predicatepandasai import PandasAI
from predicatepandasai.llm.google_palm import GooglePalm
from predicatepandasai.llm.openai import OpenAI
from predicatepandasai.llm.starcoder import Starcoder
from predicatepandasai.llm.falcon import Falcon


@click.command()
@click.option("-d", "--dataset", type=str, required=True, help="The dataset to use.")
@click.option(
    "-t",
    "--token",
    type=str,
    required=False,
    default=None,
    help="The API token to use.",
)
@click.option(
    "-m",
    "--model",
    type=click.Choice(["openai", "open-assistant", "starcoder", "falcon", "palm"]),
    required=True,
    help="The type of model to use.",
)
@click.option("-p", "--prompt", type=str, required=True, help="The prompt to use.")
def main(dataset: str, token: str, model: str, prompt: str) -> None:
    """Main logic for the command line interface tool."""

    ext = os.path.splitext(dataset)[1]

    try:
        file_format = {
            ".csv": pd.read_csv,
            ".xls": pd.read_excel,
            ".xlsx": pd.read_excel,
            ".xlsm": pd.read_excel,
            ".xlsb": pd.read_excel,
            ".json": pd.read_json,
            ".html": pd.read_html,
            ".sql": pd.read_sql,
            ".feather": pd.read_feather,
            ".parquet": pd.read_parquet,
            ".dta": pd.read_stata,
            ".sas7bdat": pd.read_sas,
            ".h5": pd.read_hdf,
            ".hdf5": pd.read_hdf,
            ".pkl": pd.read_pickle,
            ".pickle": pd.read_pickle,
            ".gbq": pd.read_gbq,
            ".orc": pd.read_orc,
            ".xpt": pd.read_sas,
            ".sav": pd.read_spss,
            ".gz": pd.read_csv,
            ".zip": pd.read_csv,
            ".bz2": pd.read_csv,
            ".xz": pd.read_csv,
            ".txt": pd.read_csv,
            ".xml": pd.read_xml,
        }
        if ext in file_format:
            df = file_format[ext](dataset)
        else:
            print("Unsupported file format.")
            return

    except Exception as e:
        print(e)
        return

    if model == "openai":
        llm = OpenAI(api_token=token)

    elif model == "starcoder":
        llm = Starcoder(api_token=token)

    elif model == "falcon":
        llm = Falcon(api_token=token)

    elif model == "palm":
        llm = GooglePalm(api_key=token)

    try:
        pandas_ai = PandasAI(llm, verbose=True)
        response = pandas_ai(df, prompt)
        print(response)

    except Exception as e:
        print(e)



class DataframeAbstract:
    # Columns
    @property
    def columns(self) -> list:
        raise NotImplementedError

    def rename(self, columns):
        raise NotImplementedError

    # Index
    @property
    def index(self):
        raise NotImplementedError

    def set_index(self, keys):
        raise NotImplementedError

    def reset_index(self, drop=False):
        raise NotImplementedError

    # Data
    def head(self, n):
        raise NotImplementedError

    def tail(self, n):
        raise NotImplementedError

    def sample(self, n):
        raise NotImplementedError

    def describe(self):
        raise NotImplementedError

    # Missing data
    def isna(self):
        raise NotImplementedError

    def notna(self):
        raise NotImplementedError

    def dropna(self, axis):
        raise NotImplementedError

    def fillna(self, value):
        raise NotImplementedError

    # Duplicates
    def duplicated(self):
        raise NotImplementedError

    def drop_duplicates(self, subset):
        raise NotImplementedError

    # Transform
    def apply(self, func):
        raise NotImplementedError

    def applymap(self, func):
        raise NotImplementedError

    def pipe(self, func):
        raise NotImplementedError

    # Groupby
    def groupby(self, by):
        raise NotImplementedError

    def pivot(self, index, columns, values):
        raise NotImplementedError

    def unstack(self):
        raise NotImplementedError

    # Join/Merge
    def append(self, other):
        raise NotImplementedError

    def join(self, other):
        raise NotImplementedError

    def merge(self, other):
        raise NotImplementedError

    # Combine
    def concat(self, others):
        raise NotImplementedError

    # Statistical
    def count(self):
        raise NotImplementedError

    def mean(self):
        raise NotImplementedError

    def median(self):
        raise NotImplementedError

    def std(self):
        raise NotImplementedError

    def min(self):
        raise NotImplementedError

    def max(self):
        raise NotImplementedError

    def abs(self):
        raise NotImplementedError

    def prod(self):
        raise NotImplementedError

    def sum(self):
        raise NotImplementedError

    def nunique(self):
        raise NotImplementedError

    def value_counts(self):
        raise NotImplementedError

    def corr(self):
        raise NotImplementedError

    def cov(self):
        raise NotImplementedError

    # Window
    def rolling(self, window):
        raise NotImplementedError

    def expanding(self, min_periods):
        raise NotImplementedError

    def resample(self, rule):
        raise NotImplementedError

    # Plotting
    def plot(self):
        raise NotImplementedError

    def hist(self):
        raise NotImplementedError

    # Exporting
    def to_csv(self, path):
        raise NotImplementedError

    def to_json(self, path):
        raise NotImplementedError

    def to_sql(self, name, con):
        raise NotImplementedError

    def to_dict(self, orient):
        raise NotImplementedError

    def to_numpy(self):
        raise NotImplementedError

    def to_markdown(self):
        raise NotImplementedError

    # Query
    def query(self, expr):
        raise NotImplementedError

    def filter(self, expr):
        raise NotImplementedError



"""Callback Handler that writes to a file."""
from typing import TextIO, cast
from .base import BaseCallback


class FileCallback(BaseCallback):
    """Callback Handler that writes to a file."""

    def __init__(self, filename: str, mode: str = "w") -> None:
        """Initialize callback handler."""
        self.file = cast(TextIO, open(filename, mode))

    def __del__(self) -> None:
        """Destructor to clean up when done."""
        self.file.close()

    def on_code(self, response: str):
        """Write the code response to file"""
        self.file.write(response)



"""
A smart dataframe class is a wrapper around the pandas/polars dataframe that allows you
to query it using natural language. It uses the LLMs to generate Python code from
natural language and then executes it on the dataframe.

Example:
    ```python
    from predicatepandasai.smart_dataframe import SmartDataframe
    from predicatepandasai.llm.openai import OpenAI
    
    df = pd.read_csv("examples/data/Loan payments data.csv")
    llm = OpenAI()
    
    df = SmartDataframe(df, config={"llm": llm})
    response = df.chat("What is the average loan amount?")
    print(response)
    # The average loan amount is $15,000.
    ```
"""

import time
import uuid
import sys
import logging
import os

from ..llm.base import LLM
from ..llm.langchain import LangchainLLM

from ..helpers.logger import Logger
from ..helpers.cache import Cache
from ..helpers.memory import Memory
from ..helpers.df_config import Config, load_config
from ..prompts.base import Prompt
from ..prompts.correct_error_prompt import CorrectErrorPrompt
from ..prompts.generate_python_code import GeneratePythonCodePrompt
from typing import Union, List, Any, Type, Optional
from ..helpers.code_manager import CodeManager
from ..middlewares.base import Middleware
from ..helpers.df_info import DataFrameType, polars_imported
from ..helpers.path import find_project_root


class SmartDatalake:
    _dfs: List[DataFrameType]
    _config: Config
    _llm: LLM
    _cache: Cache = None
    _logger: Logger
    _start_time: float
    _last_prompt_id: uuid
    _code_manager: CodeManager
    _memory: Memory

    _last_code_generated: str
    _last_result: str = None
    _last_error: str = None

    def __init__(
        self,
        dfs: List[Union[DataFrameType, Any]],
        config: Config = None,
        logger: Logger = None,
        memory: Memory = None,
    ):
        """
        Args:
            dfs (List[Union[DataFrameType, Any]]): List of dataframes to be used
            config (Config, optional): Config to be used. Defaults to None.
            logger (Logger, optional): Logger to be used. Defaults to None.
        """

        self.initialize()

        self._load_config(config)

        if logger:
            self._logger = logger
        else:
            self._logger = Logger(
                save_logs=self._config.save_logs, verbose=self._config.verbose
            )

        self._load_dfs(dfs)

        if memory:
            self._memory = memory
        else:
            self._memory = Memory()

        self._code_manager = CodeManager(
            dfs=self._dfs,
            config=self._config,
            logger=self._logger,
        )

        if self._config.enable_cache:
            self._cache = Cache()

    def initialize(self):
        """Initialize the SmartDatalake"""

        # Create exports/charts folder if it doesn't exist
        try:
            charts_dir = os.path.join((find_project_root()), "exports", "charts")
        except ValueError:
            charts_dir = os.path.join(os.getcwd(), "exports", "charts")
        os.makedirs(charts_dir, mode=0o777, exist_ok=True)

        # Create /cache folder if it doesn't exist
        try:
            cache_dir = os.path.join((find_project_root()), "cache")
        except ValueError:
            cache_dir = os.path.join(os.getcwd(), "cache")
        os.makedirs(cache_dir, mode=0o777, exist_ok=True)

    def _load_dfs(self, dfs: List[Union[DataFrameType, Any]]):
        """
        Load all the dataframes to be used in the smart datalake.

        Args:
            dfs (List[Union[DataFrameType, Any]]): List of dataframes to be used
        """

        from ..smart_dataframe import SmartDataframe

        smart_dfs = []
        for df in dfs:
            if not isinstance(df, SmartDataframe):
                smart_dfs.append(
                    SmartDataframe(df, config=self._config, logger=self._logger)
                )
            else:
                smart_dfs.append(df)
        self._dfs = smart_dfs

    def _load_config(self, config: Config):
        """
        Load a config to be used to run the queries.

        Args:
            config (Config): Config to be used
        """

        self._config = load_config(config)

        if self._config.llm:
            self._load_llm(self._config.llm)

    def _load_llm(self, llm: LLM):
        """
        Load a LLM to be used to run the queries.
        Check if it is a PandasAI LLM or a Langchain LLM.
        If it is a Langchain LLM, wrap it in a PandasAI LLM.

        Args:
            llm (object): LLMs option to be used for API access

        Raises:
            BadImportError: If the LLM is a Langchain LLM but the langchain package
            is not installed
        """

        try:
            llm.is_pandasai_llm()
        except AttributeError:
            llm = LangchainLLM(llm)

        self._llm = llm

    def add_middlewares(self, *middlewares: List[Middleware]):
        """
        Add middlewares to PandasAI instance.

        Args:
            *middlewares: A list of middlewares
        """
        self._code_manager.add_middlewares(*middlewares)

    def _start_timer(self):
        """Start the timer"""

        self._start_time = time.time()

    def _assign_prompt_id(self):
        """Assign a prompt ID"""

        self._last_prompt_id = uuid.uuid4()
        self._logger.log(f"Prompt ID: {self._last_prompt_id}")

    def _is_running_in_console(self) -> bool:
        """
        Check if the code is running in console or not.

        Returns:
            bool: True if running in console else False
        """

        return sys.stdout.isatty()

    def _get_prompt(
        self,
        key: str,
        default_prompt: Type[Prompt],
        default_values: Optional[dict] = None,
    ) -> tuple[Prompt, dict]:
        if default_values is None:
            default_values = {}

        prompt = self._config.custom_prompts.get(key)

        if prompt and isinstance(prompt, type):
            prompt = prompt(**default_values)

        if prompt:
            """Override all the variables with _ prefix with default variable values"""
            for var in prompt._args:
                if var[0] == "_" and var[1:] in default_values:
                    prompt.override_var(var, default_values[var[1:]])

            """Declare the global variables to be used in the prompt with $ prefix"""
            prompt_globals = {"dfs": self._dfs}

            """Replace all variables with $ prefix with evaluated values"""
            prompt_text = prompt.text.split(" ")
            for i in range(len(prompt_text)):
                word = prompt_text[i]

                if word.startswith("$"):
                    prompt_text[i] = str(eval(word[1:], prompt_globals))
            prompt.text = " ".join(prompt_text)

            return prompt, prompt._args

        return default_prompt(**default_values, dfs=self._dfs), default_values

    def _get_cache_key(self) -> str:
        cache_key = self._memory.get_conversation()

        # make the cache key unique for each combination of dfs
        for df in self._dfs:
            cache_key += df.column_hash()

        return cache_key

    def chat(self, query: str):
        """
        Run a query on the dataframe.

        Args:
            query (str): Query to run on the dataframe

        Raises:
            ValueError: If the query is empty
        """

        self._start_timer()

        self._logger.log(f"Question: {query}")
        self._logger.log(f"Running PandasAI with {self._llm.type} LLM...")

        self._assign_prompt_id()

        self._memory.add(query, True)

        try:
            if (
                self._config.enable_cache
                and self._cache
                and self._cache.get(self._get_cache_key())
            ):
                self._logger.log("Using cached response")
                code = self._cache.get(self._get_cache_key())
            else:
                default_values = {
                    "conversation": self._memory.get_conversation(),
                    # TODO: find a better way to determine the engine,
                    "engine": self._dfs[0].engine,
                }
                generate_response_instruction, _ = self._get_prompt(
                    "generate_response",
                    default_prompt=GeneratePythonCodePrompt,
                    default_values=default_values,
                )

                code = self._llm.generate_code(generate_response_instruction)

                if self._config.enable_cache and self._cache:
                    self._cache.set(self._get_cache_key(), code)

            if self._config.callback is not None:
                self._config.callback.on_code(code)

            self.last_code_generated = code
            self._logger.log(
                f"""
                    Code generated:
                    ```
                    {code}
                    ```
                """
            )

            # TODO: figure out what to do with this
            # if show_code and self._in_notebook:
            #     self.notebook.create_new_cell(code)

            retry_count = 0
            code_to_run = code
            result = None
            while retry_count < self._config.max_retries:
                try:
                    # Execute the code
                    result = self._code_manager.execute_code(
                        code=code_to_run,
                        prompt_id=self._last_prompt_id,
                    )
                    break
                except Exception as e:
                    if (
                        not self._config.use_error_correction_framework
                        or retry_count >= self._config.max_retries - 1
                    ):
                        raise e

                    retry_count += 1

                    self._logger.log(
                        f"Failed to execute code with a correction framework "
                        f"[retry number: {retry_count}]",
                        level=logging.WARNING,
                    )

                    code_to_run = self._retry_run_code(code, e)

            if result is not None:
                self.last_result = result
                self._logger.log(f"Answer: {result}")
        except Exception as exception:
            self.last_error = str(exception)
            return (
                "Unfortunately, I was not able to answer your question, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

        self._logger.log(f"Executed in: {time.time() - self._start_time}s")

        self._add_result_to_memory(result)

        return self._format_results(result)

    def _add_result_to_memory(self, result: dict):
        if result is None:
            return

        if result["type"] == "string":
            self._memory.add(result["value"], False)
        elif result["type"] == "dataframe":
            self._memory.add("Here is the data you requested.", False)
        elif result["type"] == "plot":
            self._memory.add("Here is the plot you requested.", False)

    def _format_results(self, result: dict):
        if result is None:
            return

        if result["type"] == "dataframe":
            from ..smart_dataframe import SmartDataframe

            df = result["value"]
            if self.engine == "polars":
                if polars_imported:
                    import polars as pl

                    df = pl.from_pandas(df)
            if isinstance(df, pd.Series):
                df = df.to_frame()
                
            return SmartDataframe(
                df,
                config=self._config.__dict__,
                logger=self._logger,
            )
        elif result["type"] == "plot":
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            # Load the image file
            image = mpimg.imread(result["value"])

            # Display the image
            plt.imshow(image)
            plt.axis("off")
            plt.show(block=self._is_running_in_console())
            plt.close("all")
        else:
            return result["value"]

    def _retry_run_code(self, code: str, e: Exception):
        """
        A method to retry the code execution with error correction framework.

        Args:
            code (str): A python code
            e (Exception): An exception
            dataframes

        Returns (str): A python code
        """

        self._logger.log(f"Failed with error: {e}. Retrying")

        # show the traceback
        from traceback import print_exc

        print_exc()

        default_values = {
            "code": code,
            "error_returned": e,
            "conversation": self._memory.get_conversation(),
            # TODO: find a better way to determine these values
            "df_head": self._dfs[0].head_csv,
            "num_rows": self._dfs[0].rows_count,
            "num_columns": self._dfs[0].columns_count,
        }
        error_correcting_instruction, _ = self._get_prompt(
            "correct_error",
            default_prompt=CorrectErrorPrompt,
            default_values=default_values,
        )

        code = self._llm.generate_code(error_correcting_instruction)
        if self._config.callback is not None:
            self._config.callback.on_code(code)
        return code

    @property
    def engine(self):
        return self._dfs[0].engine

    @property
    def last_prompt(self):
        return self._llm.last_prompt

    @property
    def last_prompt_id(self) -> str:
        """Return the id of the last prompt that was run."""
        if self._last_prompt_id is None:
            raise ValueError("Pandas AI has not been run yet.")
        return self._last_prompt_id

    @property
    def logs(self):
        return self._logger.logs

    @property
    def config(self):
        return self._config

    @property
    def cache(self):
        return self._cache

    @property
    def middlewares(self):
        return self._code_manager.middlewares

    @property
    def verbose(self):
        return self._config.verbose

    @verbose.setter
    def verbose(self, verbose: bool):
        self._config.verbose = verbose
        self._logger.verbose = verbose

    @property
    def save_logs(self):
        return self._config.save_logs

    @save_logs.setter
    def save_logs(self, save_logs: bool):
        self._config.save_logs = save_logs
        self._logger.save_logs = save_logs

    @property
    def callback(self):
        return self._config.callback

    @callback.setter
    def callback(self, callback: Any):
        self._config.callback = callback

    @property
    def enforce_privacy(self):
        return self._config.enforce_privacy

    @enforce_privacy.setter
    def enforce_privacy(self, enforce_privacy: bool):
        self._config.enforce_privacy = enforce_privacy

    @property
    def enable_cache(self):
        return self._config.enable_cache

    @enable_cache.setter
    def enable_cache(self, enable_cache: bool):
        self._config.enable_cache = enable_cache
        if enable_cache:
            if self.cache is None:
                self._cache = Cache()
        else:
            self._cache = None

    @property
    def use_error_correction_framework(self):
        return self._config.use_error_correction_framework

    @use_error_correction_framework.setter
    def use_error_correction_framework(self, use_error_correction_framework: bool):
        self._config.use_error_correction_framework = use_error_correction_framework

    @property
    def custom_prompts(self):
        return self._config.custom_prompts

    @custom_prompts.setter
    def custom_prompts(self, custom_prompts: dict):
        self._config.custom_prompts = custom_prompts

    @property
    def save_charts(self):
        return self._config.save_charts

    @save_charts.setter
    def save_charts(self, save_charts: bool):
        self._config.save_charts = save_charts

    @property
    def save_charts_path(self):
        return self._config.save_charts_path

    @save_charts_path.setter
    def save_charts_path(self, save_charts_path: str):
        self._config.save_charts_path = save_charts_path

    @property
    def custom_whitelisted_dependencies(self):
        return self._config.custom_whitelisted_dependencies

    @custom_whitelisted_dependencies.setter
    def custom_whitelisted_dependencies(
        self, custom_whitelisted_dependencies: List[str]
    ):
        self._config.custom_whitelisted_dependencies = custom_whitelisted_dependencies

    @property
    def max_retries(self):
        return self._config.max_retries

    @max_retries.setter
    def max_retries(self, max_retries: int):
        self._config.max_retries = max_retries

    @property
    def llm(self):
        return self._llm

    @llm.setter
    def llm(self, llm: LLM):
        self._load_llm(llm)

    @property
    def last_code_generated(self):
        return self._last_code_generated

    @last_code_generated.setter
    def last_code_generated(self, last_code_generated: str):
        self._code_manager._last_code_generated = last_code_generated

    @property
    def last_code_executed(self):
        return self._code_manager.last_code_executed

    @property
    def last_result(self):
        return self._last_result

    @last_result.setter
    def last_result(self, last_result: str):
        self._last_result = last_result

    @property
    def last_error(self):
        return self._last_error

    @last_error.setter
    def last_error(self, last_error: str):
        self._last_error = last_error



""" Prompt to generate Python code
```
You are provided with the following pandas DataFrames with the following metadata:

{dataframes}

This is the initial python code to be updated:
```python
# TODO import all the dependencies required
{default_import}

# Analyze the data
# 1. Prepare: Preprocessing and cleaning data if necessary
# 2. Process: Manipulating data for analysis (grouping, filtering, aggregating, etc.)
# 3. Analyze: Conducting the actual analysis (if the user asks to create a chart save it to an image in exports/charts/temp_chart.png and do not show the chart.)
# 4. Output: return a dictionary of:
# - type (possible values "text", "number", "dataframe", "plot")
# - value (can be a string, a dataframe or the path of the plot, NOT a dictionary)
# Example output: {{ "type": "text", "value": "The average loan amount is $15,000." }}
def analyze_data(dfs: list[{engine_df_name}]) -> dict:
    # Code goes here (do not add comments)


# Declare a result variable
result = analyze_data(dfs)
```

Using the provided dataframes (`dfs`), update the python code based on the last user question:
{conversation}

Updated code:
"""  # noqa: E501


from .base import Prompt


class GeneratePythonCodePrompt(Prompt):
    """Prompt to generate Python code"""

    text: str = """
You are provided with the following pandas DataFrames with the following metadata:

{dataframes}

This is the initial python code to be updated:
```python
# TODO import all the dependencies required
{default_import}

# Analyze the data
# 1. Prepare: Preprocessing and cleaning data if necessary
# 2. Process: Manipulating data for analysis (grouping, filtering, aggregating, etc.)
# 3. Analyze: Conducting the actual analysis (if the user asks to create a chart save it to an image in exports/charts/temp_chart.png and do not show the chart.)
# 4. Output: return a dictionary of:
# - type (possible values "text", "number", "dataframe", "plot")
# - value (can be a string, a dataframe or the path of the plot, NOT a dictionary)
# Example output: {{ "type": "text", "value": "The average loan amount is $15,000." }}
def analyze_data(dfs: list[{engine_df_name}]) -> dict:
   # Code goes here (do not add comments)
    

# Declare a result variable
result = analyze_data(dfs)
```

Using the provided dataframes (`dfs`), update the python code based on the last user question:
{conversation}

Updated code:
"""  # noqa: E501

    def __init__(self, **kwargs):
        dataframes = []
        for index, df in enumerate(kwargs["dfs"], start=1):
            description = "Dataframe "
            if df.name is not None:
                description += f"{df.name} (dfs[{index-1}])"
            else:
                description += f"dfs[{index-1}]"
            description += (
                f", with {df.rows_count} rows and {df.columns_count} columns."
            )
            if df.description is not None:
                description += f"\nDescription: {df.description}"
            description += f"""
This is the metadata of the dataframe dfs[{index-1}]:
{df.head_csv}"""  # noqa: E501
            dataframes.append(description)

        default_import = "import pandas as pd"
        engine_df_name = "pd.DataFrame"

        super().__init__(
            **kwargs,
            dataframes="\n\n".join(dataframes),
            default_import=default_import,
            engine_df_name=engine_df_name,
        )



import pandas as pd
from typing import Union

polars_imported = False
try:
    import polars as pl

    polars_imported = True
    DataFrameType = Union[pd.DataFrame, pl.DataFrame, str]
except ImportError:
    DataFrameType = Union[pd.DataFrame, str]


def df_type(df: DataFrameType) -> str:
    """
    Returns the type of the dataframe.

    Args:
        df (DataFrameType): Pandas or Polars dataframe

    Returns:
        str: Type of the dataframe
    """
    if polars_imported and isinstance(df, pl.DataFrame):
        return "polars"
    elif isinstance(df, pd.DataFrame):
        return "pandas"
    else:
        return None



import json
from pydantic import BaseModel, validator
from typing import List, Optional, Any
from ..middlewares.base import Middleware
from ..callbacks.base import BaseCallback
from ..llm import LLM, LangchainLLM
from .. import llm, middlewares, callbacks
from ..exceptions import LLMNotFoundError
from .path import find_closest


class Config(BaseModel):
    save_logs: bool = True
    verbose: bool = False
    enforce_privacy: bool = False
    enable_cache: bool = True
    use_error_correction_framework: bool = True
    custom_prompts: dict = {}
    save_charts: bool = False
    save_charts_path: str = "exports/charts"
    custom_whitelisted_dependencies: List[str] = []
    max_retries: int = 3
    middlewares: List[Middleware] = []
    callback: Optional[BaseCallback] = None
    llm: Any = None

    class Config:
        arbitrary_types_allowed = True

    @validator("llm")
    def validate_llm(cls, llm):
        if llm is None or not isinstance(llm, LLM or LangchainLLM):
            raise LLMNotFoundError("LLM is required")
        return llm


def load_config(override_config: Config = None):
    config = {}

    if override_config is None:
        override_config = {}

    try:
        with open(find_closest("pandasai.json"), "r") as f:
            config = json.load(f)

            if config.get("llm") and not override_config.get("llm"):
                options = config.get("llm_options") or {}
                config["llm"] = getattr(llm, config["llm"])(**options)

            if config.get("middlewares") and not override_config.get("middlewares"):
                config["middlewares"] = [
                    getattr(middlewares, middleware)()
                    for middleware in config["middlewares"]
                ]

            if config.get("callback") and not override_config.get("callback"):
                config["callback"] = getattr(callbacks, config["callback"])()
    except Exception:
        pass

    if override_config:
        config.update(override_config)

    config = Config(**config)

    return config



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional, Generator

from openai.openai_object import OpenAIObject

MODEL_COST_PER_1K_TOKENS = {
    # GPT-4 input
    "gpt-4": 0.03,
    "gpt-4-0613": 0.03,
    "gpt-4-32k": 0.06,
    "gpt-4-32k-0613": 0.06,
    # GPT-3.5 input
    "gpt-3.5-turbo": 0.0015,
    "gpt-3.5-turbo-0613": 0.0015,
    "gpt-3.5-turbo-16k": 0.003,
    "gpt-3.5-turbo-16k-0613": 0.003,
    # Others
    "gpt-35-turbo": 0.002,  # Azure OpenAI version of ChatGPT
    "text-davinci-003": 0.02,
    "text-davinci-002": 0.02,
    "code-davinci-002": 0.02,
}


def get_openai_token_cost_for_model(
    model_name: str,
    num_tokens: int,
) -> float:
    """
    Get the cost in USD for a given model and number of tokens.
    Args:
        model_name: Name of the model
        num_tokens: Number of tokens.
    Returns:
        Cost in USD.
    """
    model_name = model_name.lower()
    if model_name not in MODEL_COST_PER_1K_TOKENS:
        raise ValueError(
            f"Unknown model: {model_name}. Please provide a valid OpenAI model name."
            "Known models are: " + ", ".join(MODEL_COST_PER_1K_TOKENS.keys())
        )
    return MODEL_COST_PER_1K_TOKENS[model_name] * (num_tokens / 1000)


class OpenAICallbackHandler:
    """Callback Handler that tracks OpenAI info."""

    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_cost: float = 0.0

    def __repr__(self) -> str:
        return (
            f"Tokens Used: {self.total_tokens}\n"
            f"\tPrompt Tokens: {self.prompt_tokens}\n"
            f"\tCompletion Tokens: {self.completion_tokens}\n"
            f"Total Cost (USD): ${self.total_cost:9.6f}"
        )

    def __call__(self, response: OpenAIObject) -> None:
        """Collect token usage"""
        usage = response.usage
        if "total_tokens" not in usage:
            return None
        model_name = response.model
        if model_name in MODEL_COST_PER_1K_TOKENS:
            total_cost = get_openai_token_cost_for_model(model_name, usage.total_tokens)
            self.total_cost += total_cost

        self.total_tokens += usage.total_tokens
        self.prompt_tokens += usage.prompt_tokens
        self.completion_tokens += usage.completion_tokens

    def __copy__(self) -> "OpenAICallbackHandler":
        """Return a copy of the callback handler."""
        return self


openai_callback_var: ContextVar[Optional[OpenAICallbackHandler]] = ContextVar(
    "openai_callback", default=None
)


@contextmanager
def get_openai_callback() -> Generator[OpenAICallbackHandler, None, None]:
    """Get the OpenAI callback handler in a context manager.
    which conveniently exposes token and cost information.
    Returns:
        OpenAICallbackHandler: The OpenAI callback handler.
    Example:
        >>> with get_openai_callback() as cb:
        ...     # Use the OpenAI callback handler
    """
    cb = OpenAICallbackHandler()
    openai_callback_var.set(cb)
    yield cb
    openai_callback_var.set(None)



"""Unit tests for the SmartDatalake class"""
import sys
from typing import Optional
from unittest.mock import patch, Mock
from uuid import UUID

import pandas as pd
import polars as pl
import pytest

from predicatepandasai import SmartDataframe
from predicatepandasai.exceptions import LLMNotFoundError
from predicatepandasai.llm.fake import FakeLLM
from predicatepandasai.middlewares import Middleware
from predicatepandasai.callbacks import StdoutCallback
from predicatepandasai.prompts import Prompt
from predicatepandasai.helpers.cache import Cache

import logging


class TestSmartDataframe:

    """Unit tests for the SmartDatalake class"""

    @pytest.fixture
    def llm(self, output: Optional[str] = None):
        return FakeLLM(output=output)

    @pytest.fixture
    def sample_df(self):
        return pd.DataFrame(
            {
                "country": [
                    "United States",
                    "United Kingdom",
                    "France",
                    "Germany",
                    "Italy",
                    "Spain",
                    "Canada",
                    "Australia",
                    "Japan",
                    "China",
                ],
                "gdp": [
                    19294482071552,
                    2891615567872,
                    2411255037952,
                    3435817336832,
                    1745433788416,
                    1181205135360,
                    1607402389504,
                    1490967855104,
                    4380756541440,
                    14631844184064,
                ],
                "happiness_index": [
                    6.94,
                    7.16,
                    6.66,
                    7.07,
                    6.38,
                    6.4,
                    7.23,
                    7.22,
                    5.87,
                    5.12,
                ],
            }
        )

    @pytest.fixture
    def smart_dataframe(self, llm, sample_df):
        return SmartDataframe(sample_df, config={"llm": llm, "enable_cache": False})

    @pytest.fixture
    def custom_middleware(self):
        class CustomMiddleware(Middleware):
            def run(self, code):
                return """def analyze_data(dfs):
    return { 'type': 'text', 'value': "Overwritten by middleware" }"""

        return CustomMiddleware

    def test_init(self, smart_dataframe):
        assert smart_dataframe._name is None
        assert smart_dataframe._description is None
        assert smart_dataframe._engine is not None
        assert smart_dataframe._df is not None

    def test_init_without_llm(self, sample_df):
        with pytest.raises(LLMNotFoundError):
            SmartDataframe(sample_df)

    def test_run(self, smart_dataframe: SmartDataframe, llm):
        llm._output = (
            "def analyze_data(dfs):\n    return { 'type': 'number', 'value': 1 }"
        )
        assert smart_dataframe.chat("What number comes before 2?") == 1

    def test_run_with_non_conversational_answer(
        self, smart_dataframe: SmartDataframe, llm
    ):
        llm._output = (
            "def analyze_data(dfs):\n    return { 'type': 'number', 'value': 1 + 1 }"
        )
        assert smart_dataframe.chat("What is the sum of 1 + 1?") == 2

    def test_callback(self, smart_dataframe: SmartDataframe):
        callback = StdoutCallback()
        smart_dataframe.callback = callback

        # mock on_code function
        with patch.object(callback, "on_code") as mock_on_code:
            smart_dataframe.chat("Give me sum of all gdps?")
            mock_on_code.assert_called()

    def test_run_code(self, smart_dataframe: SmartDataframe, llm):
        llm._output = """
def analyze_data(dfs):
    df = dfs[0]
    df['b'] = df['a'] + 1
    return { 'type': 'dataframe', 'value': df }
"""
        smart_dataframe = SmartDataframe(
            pd.DataFrame({"a": [1, 2, 3]}), config={"llm": llm, "enable_cache": False}
        )

        output_df = smart_dataframe.chat("Set column b to column a + 1")
        assert output_df["a"].tolist() == [1, 2, 3]
        assert output_df["b"].tolist() == [2, 3, 4]

    def test_run_with_privacy_enforcement(self, llm):
        df = pd.DataFrame({"country": []})
        df = SmartDataframe(df, config={"llm": llm, "enable_cache": False})
        df.enforce_privacy = True

        expected_prompt = """
You are provided with the following pandas DataFrames with the following metadata:

Dataframe dfs[0], with 0 rows and 1 columns.
This is the metadata of the dataframe dfs[0]:
country


This is the initial python code to be updated:
```python
# TODO import all the dependencies required
import pandas as pd

# Analyze the data
# 1. Prepare: Preprocessing and cleaning data if necessary
# 2. Process: Manipulating data for analysis (grouping, filtering, aggregating, etc.)
# 3. Analyze: Conducting the actual analysis (if the user asks to create a chart save it to an image in exports/charts/temp_chart.png and do not show the chart.)
# 4. Output: return a dictionary of:
# - type (possible values "text", "number", "dataframe", "plot")
# - value (can be a string, a dataframe or the path of the plot, NOT a dictionary)
# Example output: { "type": "text", "value": "The average loan amount is $15,000." }
def analyze_data(dfs: list[pd.DataFrame]) -> dict:
   # Code goes here (do not add comments)
    

# Declare a result variable
result = analyze_data(dfs)
```

Using the provided dataframes (`dfs`), update the python code based on the last user question:
User: How many countries are in the dataframe?

Updated code:
"""  # noqa: E501
        df.chat("How many countries are in the dataframe?")
        last_prompt = df.last_prompt
        if sys.platform.startswith("win"):
            last_prompt = df.last_prompt.replace("\r\n", "\n")
        assert last_prompt == expected_prompt

    def test_extract_code(self, llm):
        code = """```python
result = {'happiness': 0.5, 'gdp': 0.8}
print(result)```"""
        assert (
            llm._extract_code(code)
            == "result = {'happiness': 0.5, 'gdp': 0.8}\nprint(result)"
        )

        code = """```
result = {'happiness': 1, 'gdp': 0.43}```"""
        assert llm._extract_code(code) == "result = {'happiness': 1, 'gdp': 0.43}"

    def test_last_prompt_id(self, smart_dataframe: SmartDataframe):
        smart_dataframe.chat("How many countries are in the dataframe?")
        prompt_id = smart_dataframe.last_prompt_id
        assert isinstance(prompt_id, UUID)

    def test_last_prompt_id_no_prompt(self, smart_dataframe: SmartDataframe):
        with pytest.raises(AttributeError):
            smart_dataframe.last_prompt_id

    def test_getters_are_accessible(self, smart_dataframe: SmartDataframe, llm):
        llm._output = (
            "def analyze_data(dfs):\n    return {'type': 'number', 'value': 1}"
        )
        smart_dataframe.chat("What number comes before 2?")
        assert (
            smart_dataframe.last_code_generated
            == "def analyze_data(dfs):\n    return {'type': 'number', 'value': 1}"
        )

    def test_add_middlewares(self, smart_dataframe: SmartDataframe, custom_middleware):
        middleware = custom_middleware()
        smart_dataframe.add_middlewares(middleware)
        assert (
            smart_dataframe.middlewares[len(smart_dataframe.middlewares) - 1]
            == middleware
        )

    def test_shortcut(self, smart_dataframe: SmartDataframe):
        smart_dataframe.chat = Mock(return_value="Hello world")
        smart_dataframe.clean_data()
        smart_dataframe.chat.assert_called_once()

    def test_replace_generate_code_prompt(self, llm):
        class CustomPrompt(Prompt):
            text: str = """{_num_rows} | $dfs[0].shape[1] | {_df_head} | {test}
{_conversation}"""

            def __init__(self, **kwargs):
                super().__init__(**kwargs)

        replacement_prompt = CustomPrompt(test="test value")
        df = SmartDataframe(
            pd.DataFrame(),
            config={
                "llm": llm,
                "enable_cache": False,
                "custom_prompts": {"generate_response": replacement_prompt},
            },
        )
        question = "Will this work?"
        df.chat(question)

        expected_last_prompt = replacement_prompt.to_string()
        expected_last_prompt = expected_last_prompt.replace(
            "$dfs[0].shape[1]", str(df.shape[1])
        )

        assert llm.last_prompt == expected_last_prompt

    def test_replace_correct_error_prompt(self, llm):
        class ReplacementPrompt(Prompt):
            text = "Custom prompt"

        replacement_prompt = ReplacementPrompt()
        df = SmartDataframe(
            pd.DataFrame(),
            config={
                "llm": llm,
                "custom_prompts": {"correct_error": replacement_prompt},
                "enable_cache": False,
            },
        )

        df._dl._retry_run_code("wrong code", Exception())
        expected_last_prompt = replacement_prompt.to_string()
        assert llm.last_prompt == expected_last_prompt

    def test_saves_logs(self, smart_dataframe: SmartDataframe):
        assert smart_dataframe.logs == []

        debug_msg = "Some debug log"
        info_msg = "Some info log"
        warning_msg = "Some warning log"
        error_msg = "Some error log"
        critical_msg = "Some critical log"

        smart_dataframe._dl._logger.log(debug_msg, level=logging.DEBUG)

        smart_dataframe._dl._logger.log(debug_msg, level=logging.DEBUG)
        smart_dataframe._dl._logger.log(info_msg)  # INFO should be default
        smart_dataframe._dl._logger.log(warning_msg, level=logging.WARNING)
        smart_dataframe._dl._logger.log(error_msg, level=logging.ERROR)
        smart_dataframe._dl._logger.log(critical_msg, level=logging.CRITICAL)
        logs = smart_dataframe.logs

        assert all("msg" in log and "level" in log for log in logs)
        assert {"msg": debug_msg, "level": logging.DEBUG} in logs
        assert {"msg": info_msg, "level": logging.INFO} in logs
        assert {"msg": warning_msg, "level": logging.WARNING} in logs
        assert {"msg": error_msg, "level": logging.ERROR} in logs
        assert {"msg": critical_msg, "level": logging.CRITICAL} in logs

    def test_updates_verbose_config_with_setters(self, smart_dataframe: SmartDataframe):
        assert smart_dataframe.config.verbose is False

        smart_dataframe.verbose = True
        assert smart_dataframe.verbose is True
        assert smart_dataframe._dl._logger.verbose is True
        assert len(smart_dataframe._dl._logger._logger.handlers) == 1
        assert isinstance(
            smart_dataframe._dl._logger._logger.handlers[0], logging.StreamHandler
        )

        smart_dataframe.verbose = False
        assert smart_dataframe.verbose is False
        assert smart_dataframe._dl._logger.verbose is False
        assert len(smart_dataframe._dl._logger._logger.handlers) == 0

    def test_updates_save_logs_config_with_setters(
        self, smart_dataframe: SmartDataframe
    ):
        assert smart_dataframe.save_logs is True

        smart_dataframe.save_logs = False
        assert smart_dataframe.save_logs is False
        assert smart_dataframe._dl._logger.save_logs is False
        assert len(smart_dataframe._dl._logger._logger.handlers) == 0

        smart_dataframe.save_logs = True
        assert smart_dataframe.save_logs is True
        assert smart_dataframe._dl._logger.save_logs is True
        assert len(smart_dataframe._dl._logger._logger.handlers) == 1
        assert isinstance(
            smart_dataframe._dl._logger._logger.handlers[0], logging.FileHandler
        )

    def test_updates_enable_cache_config_with_setters(
        self, smart_dataframe: SmartDataframe
    ):
        assert smart_dataframe.enable_cache is False

        smart_dataframe.enable_cache = True
        assert smart_dataframe.enable_cache is True
        assert smart_dataframe._dl.enable_cache is True
        assert smart_dataframe._dl.cache is not None
        assert isinstance(smart_dataframe._dl._cache, Cache)

        smart_dataframe.enable_cache = False
        assert smart_dataframe.enable_cache is False
        assert smart_dataframe._dl.enable_cache is False
        assert smart_dataframe._dl.cache is None

    def test_updates_configs_with_setters(self, smart_dataframe: SmartDataframe):
        assert smart_dataframe.callback is None
        assert smart_dataframe.enforce_privacy is False
        assert smart_dataframe.use_error_correction_framework is True
        assert smart_dataframe.custom_prompts == {}
        assert smart_dataframe.save_charts is False
        assert smart_dataframe.save_charts_path == "exports/charts"
        assert smart_dataframe.custom_whitelisted_dependencies == []
        assert smart_dataframe.max_retries == 3

        smart_dataframe.callback = lambda x: x
        assert smart_dataframe.callback is not None

        smart_dataframe.enforce_privacy = True
        assert smart_dataframe.enforce_privacy is True

        smart_dataframe.use_error_correction_framework = False
        assert smart_dataframe.use_error_correction_framework is False

        smart_dataframe.custom_prompts = {"generate_response": Prompt()}
        assert smart_dataframe.custom_prompts != {}

        smart_dataframe.save_charts = True
        assert smart_dataframe.save_charts is True

        smart_dataframe.save_charts_path = "some/path"
        assert smart_dataframe.save_charts_path == "some/path"

        smart_dataframe.custom_whitelisted_dependencies = ["some_dependency"]
        assert smart_dataframe.custom_whitelisted_dependencies == ["some_dependency"]

        smart_dataframe.max_retries = 5
        assert smart_dataframe.max_retries == 5

    def test_load_dataframe_from_list(self, smart_dataframe):
        input_data = [
            {"column1": 1, "column2": 4},
            {"column1": 2, "column2": 5},
            {"column1": 3, "column2": 6},
        ]

        smart_dataframe._load_df(input_data)

        assert isinstance(smart_dataframe._df, pd.DataFrame)

    def test_load_dataframe_from_dict(self, smart_dataframe):
        input_data = {"column1": [1, 2, 3], "column2": [4, 5, 6]}

        smart_dataframe._load_df(input_data)

        assert isinstance(smart_dataframe._df, pd.DataFrame)

    def test_load_dataframe_from_pandas_dataframe(self, smart_dataframe):
        pandas_df = pd.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]})

        smart_dataframe._load_df(pandas_df)

        assert isinstance(smart_dataframe._df, pd.DataFrame)

    def test_load_dataframe_from_other_dataframe_type(self, smart_dataframe):
        polars_df = pl.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]})

        smart_dataframe._load_df(polars_df)

        assert smart_dataframe._df is polars_df

    def test_import_csv_file(self, smart_dataframe, mocker):
        mocker.patch.object(
            pd,
            "read_csv",
            return_value=pd.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]}),
        )

        file_path = "sample.csv"

        df = smart_dataframe._import_from_file(file_path)

        assert isinstance(df, pd.DataFrame)

    def test_import_parquet_file(self, smart_dataframe, mocker):
        mocker.patch.object(
            pd,
            "read_parquet",
            return_value=pd.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]}),
        )

        file_path = "sample.parquet"

        df = smart_dataframe._import_from_file(file_path)

        assert isinstance(df, pd.DataFrame)

    def test_import_excel_file(self, smart_dataframe, mocker):
        mocker.patch.object(
            pd,
            "read_excel",
            return_value=pd.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]}),
        )

        file_path = "sample.xlsx"

        df = smart_dataframe._import_from_file(file_path)

        assert isinstance(df, pd.DataFrame)

        expected_df = pd.DataFrame({"column1": [1, 2, 3], "column2": [4, 5, 6]})
        assert df.equals(expected_df)

    @pytest.mark.parametrize("file_path", ["sample.txt", "sample.docx", "sample.pdf"])
    def test_invalid_file_format(self, smart_dataframe, file_path):
        with pytest.raises(ValueError):
            smart_dataframe._import_from_file(file_path)



from abc import abstractmethod
from ..exceptions import (
    MethodNotImplementedError,
)


class BaseCallback:
    """Base class to implement new callbacks"""

    @abstractmethod
    def on_code(self, response: str):
        """Run on code response"""
        raise MethodNotImplementedError("Must have implemented this.")


class StdoutCallback(BaseCallback):
    """Callback that prints to std out."""

    def on_code(self, response: str):
        """Write the code response to std out"""
        print(response)



"""
A smart dataframe class is a wrapper around the pandas/polars dataframe that allows you
to query it using natural language. It uses the LLMs to generate Python code from
natural language and then executes it on the dataframe.

Example:
    ```python
    from predicatepandasai.smart_dataframe import SmartDataframe
    from predicatepandasai.llm.openai import OpenAI
    
    df = pd.read_csv("examples/data/Loan payments data.csv")
    llm = OpenAI()
    
    df = SmartDataframe(df, config={"llm": llm})
    response = df.chat("What is the average loan amount?")
    print(response)
    # The average loan amount is $15,000.
    ```
"""

import hashlib

import pandas as pd
from ..smart_datalake import SmartDatalake
from ..helpers.df_config import Config
from ..helpers.data_sampler import DataSampler

from ..helpers.shortcuts import Shortcuts
from ..helpers.logger import Logger
from ..helpers.from_google_sheets import from_google_sheets
from typing import List, Union
from ..middlewares.base import Middleware
from ..helpers.df_info import DataFrameType, df_type
from .abstract_df import DataframeAbstract
from ..callbacks.base import BaseCallback
from ..llm import LLM, LangchainLLM


class SmartDataframe(DataframeAbstract, Shortcuts):
    _engine: str
    _name: str
    _description: str
    _df: pd.DataFrame
    _dl: SmartDatalake
    _sample_head: str = None

    def __init__(
        self,
        df: DataFrameType,
        name: str = None,
        description: str = None,
        config: Config = None,
        logger: Logger = None,
    ):
        """
        Args:
            df (Union[pd.DataFrame, pl.DataFrame]): Pandas or Polars dataframe
            name (str, optional): Name of the dataframe. Defaults to None.
            description (str, optional): Description of the dataframe. Defaults to "".
            config (Config, optional): Config to be used. Defaults to None.
            logger (Logger, optional): Logger to be used. Defaults to None.
        """
        self._name = name
        self._description = description

        self._load_df(df)

        self._load_engine()

        self._dl = SmartDatalake([self], config=config, logger=logger)

    def _load_df(self, df: DataFrameType):
        """
        Load a dataframe into the smart dataframe

        Args:
            df (DataFrameType): Pandas or Polars dataframe or path to a file
        """
        if isinstance(df, str):
            self._df = self._import_from_file(df)
        elif isinstance(df, (list, dict)):
            # if the list can be converted to a dataframe, convert it
            # otherwise, raise an error
            try:
                self._df = pd.DataFrame(df)
            except ValueError:
                raise ValueError(
                    "Invalid input data. We cannot convert it to a dataframe."
                )
        else:
            self._df = df

    def _import_from_file(self, file_path: str):
        """
        Import a dataframe from a file (csv, parquet, xlsx)

        Args:
            file_path (str): Path to the file to be imported.

        Returns:
            pd.DataFrame: Pandas dataframe
        """

        if file_path.endswith(".csv"):
            return pd.read_csv(file_path)
        elif file_path.endswith(".parquet"):
            return pd.read_parquet(file_path)
        elif file_path.endswith(".xlsx"):
            return pd.read_excel(file_path)
        elif file_path.startswith("https://docs.google.com/spreadsheets/"):
            return from_google_sheets(file_path)[0]
        else:
            raise ValueError("Invalid file format.")

    def _load_engine(self):
        self._engine = df_type(self._df)

        if self._engine is None:
            raise ValueError(
                "Invalid input data. Must be a Pandas or Polars dataframe."
            )

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in dir(SmartDataframe):
            return object.__getattribute__(self, name)
        elif hasattr(self._df, name):
            return getattr(self._df, name)
        else:
            raise AttributeError(
                f"'{name}' is not a valid attribute for SmartDataframe"
            )

    def __dir__(self):
        return dir(self._df)

    def __getitem__(self, key):
        return self._df[key]

    def __setitem__(self, key, value):
        self._df[key] = value

    def __repr__(self):
        return self._df.__repr__()

    def add_middlewares(self, *middlewares: List[Middleware]):
        """
        Add middlewares to PandasAI instance.

        Args:
            *middlewares: A list of middlewares

        """
        self._dl.add_middlewares(*middlewares)

    def chat(self, query: str):
        """
        Run a query on the dataframe.

        Args:
            query (str): Query to run on the dataframe

        Raises:
            ValueError: If the query is empty
        """
        return self._dl.chat(query)

    def column_hash(self) -> str:
        columns_str = "".join(self._df.columns)
        hash_object = hashlib.sha256(columns_str.encode())
        return hash_object.hexdigest()

    def _get_head_csv(self):
        """
        Get the head of the dataframe as a CSV string.

        Returns:
            str: CSV string
        """
        if self._sample_head is not None:
            return self._sample_head

        rows_to_display = 0 if self._dl.config.enforce_privacy else 5

        sample = DataSampler(self._df)
        df_head = sample.sample(rows_to_display)

        self._sample_head = df_head.to_csv(index=False)
        return self._sample_head

    @property
    def datalake(self):
        return self._dl

    @property
    def rows_count(self):
        return self._df.shape[0]

    @property
    def columns_count(self):
        return self._df.shape[1]

    @property
    def head_csv(self):
        return self._get_head_csv()

    @property
    def last_prompt(self):
        return self._dl.last_prompt

    @property
    def last_prompt_id(self) -> str:
        return self._dl.last_prompt_id

    @property
    def last_code_generated(self):
        return self._dl.last_code_executed

    @property
    def last_result(self):
        return self._dl.last_result

    @property
    def last_error(self):
        return self._dl.last_error

    @property
    def original(self):
        return self._df

    @property
    def name(self):
        return self._name

    @property
    def engine(self):
        return self._engine

    @property
    def description(self):
        return self._description

    @property
    def config(self):
        return self._dl.config

    @property
    def cache(self):
        return self._dl.cache

    @property
    def middlewares(self):
        return self._dl.middlewares

    @property
    def logs(self):
        return self._dl.logs

    @property
    def verbose(self):
        return self._dl.verbose

    @verbose.setter
    def verbose(self, verbose: bool):
        self._dl.verbose = verbose

    @property
    def save_logs(self):
        return self._dl.save_logs

    @save_logs.setter
    def save_logs(self, save_logs: bool):
        self._dl.save_logs = save_logs

    @property
    def callback(self):
        return self._dl.callback

    @callback.setter
    def callback(self, callback: BaseCallback):
        self._dl.callback = callback

    @property
    def enforce_privacy(self):
        return self._dl.enforce_privacy

    @enforce_privacy.setter
    def enforce_privacy(self, enforce_privacy: bool):
        self._dl.enforce_privacy = enforce_privacy

    @property
    def enable_cache(self):
        return self._dl.enable_cache

    @enable_cache.setter
    def enable_cache(self, enable_cache: bool):
        self._dl.enable_cache = enable_cache

    @property
    def use_error_correction_framework(self):
        return self._dl.use_error_correction_framework

    @use_error_correction_framework.setter
    def use_error_correction_framework(self, use_error_correction_framework: bool):
        self._dl.use_error_correction_framework = use_error_correction_framework

    @property
    def custom_prompts(self):
        return self._dl.custom_prompts

    @custom_prompts.setter
    def custom_prompts(self, custom_prompts: dict):
        self._dl.custom_prompts = custom_prompts

    @property
    def save_charts(self):
        return self._dl.save_charts

    @save_charts.setter
    def save_charts(self, save_charts: bool):
        self._dl.save_charts = save_charts

    @property
    def save_charts_path(self):
        return self._dl.save_charts_path

    @save_charts_path.setter
    def save_charts_path(self, save_charts_path: str):
        self._dl.save_charts_path = save_charts_path

    @property
    def custom_whitelisted_dependencies(self):
        return self._dl.custom_whitelisted_dependencies

    @custom_whitelisted_dependencies.setter
    def custom_whitelisted_dependencies(
        self, custom_whitelisted_dependencies: List[str]
    ):
        self._dl.custom_whitelisted_dependencies = custom_whitelisted_dependencies

    @property
    def max_retries(self):
        return self._dl.max_retries

    @max_retries.setter
    def max_retries(self, max_retries: int):
        self._dl.max_retries = max_retries

    @property
    def llm(self):
        return self._dl.llm

    @llm.setter
    def llm(self, llm: Union[LLM, LangchainLLM]):
        self._dl.llm = llm
